"""Baseclass for unary operators."""
from .. import evaluation, approximation
from ..baseclass import Dist


class UnaryOperator(Dist):

    def _precedence_order(self):
        """Precedence order of the various dimensions."""
        dist = self.prm["dist"]
        if isinstance(dist, Dist):
            indices = dist._precedence_order()
        else:
            indices = list(range(len(self)))
        return indices

    def _pdf(self, xloc, dist, cache, **kwargs):
        """Probability density function."""
        xloc_ = self._pre_fwd(xloc, **kwargs)
        qloc = evaluation.evaluate_density(dist, xloc_, cache=cache)
        qloc *= self._post_pdf(xloc, **kwargs)
        return qloc

    def _cdf(self, xloc, dist, cache, **kwargs):
        """Cumulative distribution function."""
        xloc = self._pre_fwd(xloc, **kwargs)
        uloc = evaluation.evaluate_forward(dist, xloc, cache=cache)
        return self._post_fwd(uloc, **kwargs)

    def _ppf(self, q, dist, cache, **kwargs):
        """Point percentile function."""
        q = self._pre_inv(q, **kwargs)
        uloc = evaluation.evaluate_inverse(dist, q, cache=cache)
        return self._post_inv(uloc, **kwargs)

    def _lower(self, dist, cache, **kwargs):
        """Distribution bounds."""
        uloc = evaluation.evaluate_lower(dist, cache=cache)
        return self._post_inv(uloc, **kwargs)

    def _upper(self, dist, cache, **kwargs):
        """Distribution bounds."""
        uloc = evaluation.evaluate_upper(dist, cache=cache)
        return self._post_inv(uloc, **kwargs)

    def _mom(self, x, dist, cache, **kwargs):
        return approximation.approximate_moment(self, x)

    def __len__(self):
        return len(self.prm["dist"])

    def _fwd_cache(self, cache):
        kwargs = {key: evaluation.get_forward_cache(value, cache)
                  for key, value in self.prm.items()}
        if not any(isinstance(value, Dist) for value in kwargs.values()):
            return self._post_inv(kwargs.pop("dist"), **kwargs)
        return self

    def _inv_cache(self, cache):
        kwargs = {key: evaluation.get_forward_cache(value, cache)
                  for key, value in self.prm.items()}
        if not any(isinstance(value, Dist) for value in kwargs.values()):
            return self._pre_fwd(kwargs.pop("dist"), **kwargs)
        return self



"""Pareto type 2 distribution."""
from ..baseclass import Dist
from ..operators.addition import Add


class pareto2(Dist):
    """Pareto type 2 distribution."""

    def __init__(self, c):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return c*1.0/(1.0+x)**(c+1.0)

    def _cdf(self, x, c):
        return 1.0-1.0/(1.0+x)**c

    def _ppf(self, q, c):
        return pow(1.0-q,-1.0/c)-1

    def _lower(self, c):
        return 0.0


class Pareto2(Add):
    """
    Pareto type 2 distribution.

    Also known as Lomax distribution (for loc=0).

    Lower threshold at loc and survival: (1+x)^-shape.

    Args:
        shape (float, Dist):
            Shape parameter
        scale (float, Dist):
            Scaling parameter
        loc (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.Pareto2(2, 2, 2)
        >>> distribution
        Pareto2(loc=2, scale=2, shape=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([2.2361, 2.582 , 3.1623, 4.4721])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.7155, 0.4648, 0.253 , 0.0894])
        >>> distribution.sample(4).round(4)
        array([3.3981, 2.126 , 8.9697, 2.7794])
    """

    def __init__(self, shape=1, scale=1, loc=0):
        self._repr = {"shape": shape, "scale": scale, "loc": loc}
        Add.__init__(self, left=pareto2(shape)*scale, right=loc)



"""Anglit distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add


class anglit(Dist):
    """Anglit distribution."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return numpy.cos(2*x)

    def _cdf(self, x):
        return numpy.sin(x+numpy.pi/4)**2.0

    def _ppf(self, q):
        return (numpy.arcsin(numpy.sqrt(q))-numpy.pi/4)

    def _lower(self):
        return -numpy.pi/4

    def _upper(self):
        return numpy.pi/4


class Anglit(Add):
    """
    Anglit distribution.

    Args:
        loc (float, Dist):
            Location parameter
        scale (float, Dist):
            Scaling parameter

    Examples:
        >>> distribution = chaospy.Anglit(2, 4)
        >>> distribution
        Anglit(loc=2, scale=4)
        >>> q = numpy.linspace(0, 1, 5)
        >>> distribution.inv(q).round(4)
        array([-1.1416,  0.9528,  2.    ,  3.0472,  5.1416])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.  , 0.25, 0.5 , 0.75, 1.  ])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.    , 0.2165, 0.25  , 0.2165, 0.    ])
        >>> distribution.sample(4).round(4)
        array([2.6245, 0.2424, 4.2421, 1.9288])
        >>> distribution.mom([1, 2, 3]).round(4)
        array([ 2.    ,  5.8696, 19.2176])
    """

    def __init__(self, loc=0, scale=1):
        self._repr = {"scale": scale, "loc": loc}
        Add.__init__(self, left=anglit()*scale, right=loc)



"""Gamma distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class gamma(Dist):

    def __init__(self, a=1):
        Dist.__init__(self, a=a)

    def _pdf(self, x, a):
        return x**(a-1)*numpy.e**(-x) / special.gamma(a)

    def _cdf(self, x, a):
        return special.gammainc(a, x)

    def _ppf(self, q, a):
        return special.gammaincinv(a, q)

    def _mom(self, k, a):
        return special.gamma(a+k)/special.gamma(a)

    def _ttr(self, n, a):
        return 2.*n+a, n*n+n*(a-1)

    def _lower(self, a):
        return 0.

    def _upper(self, a):
        return 40+2*a


class Gamma(Add):
    """
    Gamma distribution.

    Also an Erlang distribution when shape=k and scale=1./lamb.

    Args:
        shape (float, Dist):
            Shape parameter. a>0.
        scale (float, Dist):
            Scale parameter. scale!=0
        shift (float, Dist):
            Location of the lower bound.

    Examples:
        >>> distribution = chaospy.Gamma(1, 1, 1)
        >>> distribution
        Gamma(scale=1, shape=1, shift=1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([1.2231, 1.5108, 1.9163, 2.6094])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.8, 0.6, 0.4, 0.2])
        >>> distribution.sample(4).round(4)
        array([2.0601, 1.1222, 4.0014, 1.6581])
        >>> distribution.mom(1)
        array(2.)
        >>> distribution.ttr([1, 2, 3]).round(4)
        array([[4., 6., 8.],
               [1., 4., 9.]])
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=gamma(shape)*scale, right=shift)


class Exponential(Add):
    R"""
    Exponential Probability Distribution

    Args:
        scale (float, Dist):
            Scale parameter. scale!=0
        shift (float, Dist):
            Location of the lower bound.

    Examples;:
        >>> distribution = chaospy.Exponential(2, 3)
        >>> distribution
        Exponential(scale=2, shift=3)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([3.4463, 4.0217, 4.8326, 6.2189])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.sample(4).round(4)
        array([5.1203, 3.2444, 9.0028, 4.3163])
        >>> distribution.mom(1).round(4)
        5.0
        >>> distribution.ttr([1, 2, 3]).round(4)
        array([[ 9., 13., 17.],
               [ 4., 16., 36.]])
    """

    def __init__(self, scale=1, shift=0):
        self._repr = {"scale": scale, "shift": shift}
        Add.__init__(self, left=gamma(1)*scale, right=shift)



"""Levy distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class levy(Dist):
    """Levy distribution."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return 1/numpy.sqrt(2*numpy.pi*x)/x*numpy.exp(-1/(2*x))

    def _cdf(self, x):
        return 2*(1-special.ndtr(1/numpy.sqrt(x)))

    def _ppf(self, q):
        val = special.ndtri(1-q/2.0)
        return 1.0/(val*val)

    def _lower(self):
        return 0.


class Levy(Add):
    """
    Levy distribution

    Args:
        loc (float, Dist):
            Location parameter
        scale (float, Dist):
            Scaling parameter

    Examples:
        >>> distribution = chaospy.Levy(2, 2)
        >>> distribution
        Levy(loc=2, scale=2)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([ 3.2177,  4.8236,  9.2728, 33.16  ])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.1847, 0.0834, 0.0251, 0.0031])
        >>> distribution.sample(4).round(4)
        array([ 11.9303,   2.8051, 516.4406,   6.0494])
    """
    def __init__(self, loc=0, scale=1):
        self._repr = {"loc": loc, "scale": scale}
        Add.__init__(self, left=levy()*scale, right=loc)



"""Generalized exponential distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class generalized_exponential(Dist):
    """Generalized exponential distribution."""

    def __init__(self, a=1, b=1, c=1):
        Dist.__init__(self, a=a, b=b, c=c)

    def _pdf(self, x, a, b, c):
        return (a+b*(-numpy.expm1(-c*x)))*numpy.exp((-a-b)*x+b*(-numpy.expm1(-c*x))/c)

    def _cdf(self, x, a, b, c):
        output = -numpy.expm1((-a-b)*x + b*(-numpy.expm1(-c*x))/c)
        output = numpy.where(x > 0, output, 0)
        return output

    def _lower(self, a, b, c):
        return 0.

    def _upper(self, a, b, c):
        return 10**10


class GeneralizedExponential(Add):
    """
    Generalized exponential distribution.

    Args:
        a (float, Dist):
            First shape parameter
        b (float, Dist):
            Second shape parameter
        c (float, Dist):
            Third shape parameter
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Note:
        "An Extension of Marshall and Olkin's Bivariate Exponential Distribution",
        H.K. Ryu, Journal of the American Statistical Association, 1993.

        "The Exponential Distribution: Theory, Methods and Applications",
        N. Balakrishnan, Asit P. Basu.

    Examples:
        >>> distribution = chaospy.GeneralizedExponential(3, 2, 2, 2, 2)
        >>> distribution
        GeneralizedExponential(a=3, b=2, c=2, scale=2, shift=2)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([2.1423, 2.3113, 2.5314, 2.8774])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([1.3061, 1.0605, 0.7649, 0.4168])
        >>> distribution.sample(4).round(4)
        array([3.3106, 2.0498, 3.3575, 2.7079])
    """

    def __init__(self, a=1, b=1, c=1, scale=1, shift=0):
        self._repr = {"a": a, "b": b, "c": c, "scale": scale, "shift": shift}
        Add.__init__(
            self, left=generalized_exponential(a, b, c)*scale, right=shift)



"""Multivariate Log-Normal Distribution."""
import numpy
from scipy import special

from ..baseclass import Dist


class MvLogNormal(Dist):
    """
    Multivariate Log-Normal Distribution.

    Args:
        loc (float, Dist):
            Mean vector
        scale (float, Dist):
            Covariance matrix or variance vector if scale is a 1-d vector.

    Examples:
        >>> distribution = chaospy.MvLogNormal([1, 2], [[1, 0.6], [0.6, 1]])
        >>> distribution
        MvLogNormal(loc=[1.0, 2.0], scale=[[1.0, 0.6], [0.6, 1.0]])
        >>> mesh = numpy.meshgrid(*[numpy.linspace(0, 1, 5)[1:-1]]*2)
        >>> distribution.inv(mesh).round(4)
        array([[[ 1.3847,  2.7183,  5.3361],
                [ 1.3847,  2.7183,  5.3361],
                [ 1.3847,  2.7183,  5.3361]],
        <BLANKLINE>
               [[ 2.874 ,  4.3077,  6.4566],
                [ 4.9298,  7.3891, 11.075 ],
                [ 8.4562, 12.6745, 18.9971]]])
        >>> distribution.fwd(distribution.inv(mesh)).round(4)
        array([[[0.25, 0.5 , 0.75],
                [0.25, 0.5 , 0.75],
                [0.25, 0.5 , 0.75]],
        <BLANKLINE>
               [[0.25, 0.25, 0.25],
                [0.5 , 0.5 , 0.5 ],
                [0.75, 0.75, 0.75]]])
        >>> distribution.pdf(distribution.inv(mesh)).round(4)
        array([[0.0317, 0.0135, 0.0037],
               [0.0232, 0.0099, 0.0027],
               [0.0108, 0.0046, 0.0012]])
        >>> distribution.sample(4).round(4)
        array([[ 4.0351,  0.8185, 14.1201,  2.5996],
               [23.279 ,  1.8986,  4.9261,  5.8399]])
        >>> distribution.mom((1, 2)).round(4)
        6002.9122
    """

    def __init__(self, loc=[0, 0], scale=[[1, 0.5], [0.5, 1]]):
        loc = numpy.asfarray(loc)
        scale = numpy.asfarray(scale)
        assert len(loc) == len(scale)
        self._repr = {"loc": str(loc.tolist()), "scale": str(scale.tolist())}

        C = numpy.linalg.cholesky(scale)
        Ci = numpy.linalg.inv(C)
        Dist.__init__(self, loc=loc, C=C, Ci=Ci, scale=scale)

    def _cdf(self, x, loc, C, Ci, scale):
        y = numpy.log(numpy.abs(x) + 1.*(x<=0))
        out = special.ndtr(numpy.dot(Ci, (y.T-loc.T).T))
        return numpy.where(x <= 0, 0., out)

    def _ppf(self, q, loc, C, Ci, scale):
        return numpy.e**(numpy.dot(C, special.ndtri(q)).T+loc.T).T

    def _mom(self, k, loc, C, Ci, scale):
        output =  numpy.dot(k, loc)
        output += .5*numpy.dot(numpy.dot(k, scale), k)
        output =  numpy.e**(output)
        return output

    def _lower(self, loc, C, Ci, scale):
        return 0.

    def _upper(self, loc, C, Ci, scale):
        return numpy.e**(7.1*numpy.sqrt(numpy.diag(scale)) + loc.T).T

    def __len__(self):
        return len(self.prm["C"])



"""
Addition operator.

Example usage
-------------

Distribution and a constant::

    >>> distribution = chaospy.Normal(0, 1) + 10
    >>> distribution
    Add(Normal(mu=0, sigma=1), 10)
    >>> distribution.sample(5).round(4)
    array([10.395 ,  8.7997, 11.6476,  9.9553, 11.1382])
    >>> distribution.fwd([9, 10, 11]).round(4)
    array([0.1587, 0.5   , 0.8413])
    >>> distribution.inv(distribution.fwd([9, 10, 11])).round(4)
    array([ 9., 10., 11.])
    >>> distribution.pdf([9, 10, 11]).round(4)
    array([0.242 , 0.3989, 0.242 ])
    >>> distribution.mom([1, 2, 3]).round(4)
    array([  10.,  101., 1030.])
    >>> distribution.ttr([1, 2, 3]).round(4)
    array([[10., 10., 10.],
           [ 1.,  2.,  3.]])

Construct joint addition distribution::

    >>> lhs = chaospy.Uniform(2, 3)
    >>> rhs = chaospy.Uniform(3, 4)
    >>> addition = lhs + rhs
    >>> addition
    Add(Uniform(lower=2, upper=3), Uniform(lower=3, upper=4))
    >>> joint1 = chaospy.J(lhs, addition)
    >>> joint2 = chaospy.J(rhs, addition)

Generate random samples::

    >>> joint1.sample(4).round(4)
    array([[2.2123, 2.0407, 2.3972, 2.2331],
           [6.0541, 5.2478, 6.1397, 5.6253]])
    >>> joint2.sample(4).round(4)
    array([[3.1823, 3.7435, 3.0696, 3.8853],
           [6.1349, 6.6747, 5.485 , 5.9143]])

Forward transformations::

    >>> lcorr = numpy.array([2.1, 2.5, 2.9])
    >>> rcorr = numpy.array([3.01, 3.5, 3.99])
    >>> joint1.fwd([lcorr, lcorr+rcorr]).round(4)
    array([[0.1 , 0.5 , 0.9 ],
           [0.01, 0.5 , 0.99]])
    >>> joint2.fwd([rcorr, lcorr+rcorr]).round(4)
    array([[0.01, 0.5 , 0.99],
           [0.1 , 0.5 , 0.9 ]])

Inverse transformations::

    >>> joint1.inv(joint1.fwd([lcorr, lcorr+rcorr])).round(4)
    array([[2.1 , 2.5 , 2.9 ],
           [5.11, 6.  , 6.89]])
    >>> joint2.inv(joint2.fwd([rcorr, lcorr+rcorr])).round(4)
    array([[3.01, 3.5 , 3.99],
           [5.11, 6.  , 6.89]])

Raw moments::

    >>> joint1.mom([(0, 1, 1), (1, 0, 1)]).round(4)
    array([ 6.0006,  2.5002, 15.0847])
    >>> joint2.mom([(0, 1, 1), (1, 0, 1)]).round(4)
    array([ 6.0006,  3.5003, 21.0853])
"""
from __future__ import division
from scipy.special import comb
import numpy

from ..baseclass import Dist, StochasticallyDependentError
from .. import evaluation
from .binary import BinaryOperator


class Add(BinaryOperator):
    """Addition."""

    def _lower(self, left, right, cache):
        """
        Distribution bounds.

        Example:
            >>> chaospy.Uniform().lower
            array([0.])
            >>> chaospy.Add(chaospy.Uniform(), 2).lower
            array([2.])
            >>> chaospy.Add(2, chaospy.Uniform()).lower
            array([2.])
            >>> chaospy.Add(1, 1).lower
            array([2.])
        """
        left = evaluation.get_forward_cache(left, cache)
        right = evaluation.get_forward_cache(right, cache)
        if isinstance(left, Dist):
            left = evaluation.evaluate_lower(left, cache=cache)
        if isinstance(right, Dist):
            right = evaluation.evaluate_lower(right, cache=cache)
        return left+right

    def _upper(self, left, right, cache):
        """
        Distribution bounds.

        Example:
            >>> chaospy.Uniform().upper
            array([1.])
            >>> chaospy.Add(chaospy.Uniform(), 2).upper
            array([3.])
            >>> chaospy.Add(2, chaospy.Uniform()).upper
            array([3.])
            >>> chaospy.Add(1, 1).upper
            array([2.])
        """
        left = evaluation.get_forward_cache(left, cache)
        right = evaluation.get_forward_cache(right, cache)
        if isinstance(left, Dist):
            left = evaluation.evaluate_upper(left, cache=cache)
        if isinstance(right, Dist):
            right = evaluation.evaluate_upper(right, cache=cache)
        return left+right

    def _pre_fwd_left(self, xloc, other):
        xloc = (xloc.T-numpy.asfarray(other).T).T
        return xloc

    def _pre_fwd_right(self, xloc, other):
        xloc = (xloc.T-numpy.asfarray(other).T).T
        return xloc

    def _post_fwd(self, uloc, other):
        return uloc

    def _alt_fwd(self, xloc, left, right):
        return numpy.asfarray(left+right <= xloc)

    def _pdf(self, xloc, left, right, cache):
        """
        Probability density function.

        Example:
            >>> print(chaospy.Uniform().pdf([-2, 0, 2, 4]))
            [0. 1. 0. 0.]
            >>> print(chaospy.Add(chaospy.Uniform(), 2).pdf([-2, 0, 2, 4]))
            [0. 0. 1. 0.]
            >>> print(chaospy.Add(2, chaospy.Uniform()).pdf([-2, 0, 2, 4]))
            [0. 0. 1. 0.]
            >>> print(chaospy.Add(1, 1).pdf([-2, 0, 2, 4])) # Dirac logic
            [ 0.  0. inf  0.]
        """
        left = evaluation.get_forward_cache(left, cache)
        right = evaluation.get_forward_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise evaluation.DependencyError(
                    "under-defined distribution {} or {}".format(left, right))
        elif not isinstance(right, Dist):
            return numpy.inf
        else:
            left, right = right, left

        xloc = (xloc.T-numpy.asfarray(right).T).T
        output = evaluation.evaluate_density(left, xloc, cache=cache)
        assert output.shape == xloc.shape
        return output

    def _ppf(self, uloc, left, right, cache):
        """
        Point percentile function.

        Example:
            >>> print(chaospy.Uniform().inv([0.1, 0.2, 0.9]))
            [0.1 0.2 0.9]
            >>> print(chaospy.Add(chaospy.Uniform(), 2).inv([0.1, 0.2, 0.9]))
            [2.1 2.2 2.9]
            >>> print(chaospy.Add(2, chaospy.Uniform()).inv([0.1, 0.2, 0.9]))
            [2.1 2.2 2.9]
            >>> print(chaospy.Add(1, 1).inv([0.1, 0.2, 0.9]))
            [2. 2. 2.]
        """
        left = evaluation.get_inverse_cache(left, cache)
        right = evaluation.get_inverse_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise evaluation.DependencyError(
                    "under-defined distribution {} or {}".format(left, right))
        elif not isinstance(right, Dist):
            return left+right
        else:
            left, right = right, left

        xloc = evaluation.evaluate_inverse(left, uloc, cache=cache)
        output = (xloc.T + numpy.asfarray(right).T).T
        return output

    def _mom(self, keys, left, right, cache):
        """
        Statistical moments.

        Example:
            >>> print(numpy.around(chaospy.Uniform().mom([0, 1, 2, 3]), 4))
            [1.     0.5    0.3333 0.25  ]
            >>> print(numpy.around(chaospy.Add(chaospy.Uniform(), 2).mom([0, 1, 2, 3]), 4))
            [ 1.      2.5     6.3333 16.25  ]
            >>> print(numpy.around(chaospy.Add(2, chaospy.Uniform()).mom([0, 1, 2, 3]), 4))
            [ 1.      2.5     6.3333 16.25  ]
            >>> print(numpy.around(chaospy.Add(1, 1).mom([0, 1, 2, 3]), 4))
            [1. 2. 4. 8.]
        """
        if evaluation.get_dependencies(left, right):
            raise evaluation.DependencyError(
                "sum of dependent distributions not feasible: "
                "{} and {}".format(left, right)
            )

        keys_ = numpy.mgrid[tuple(slice(0, key+1, 1) for key in keys)]
        keys_ = keys_.reshape(len(self), -1)

        if isinstance(left, Dist):
            left = [
                evaluation.evaluate_moment(left, key, cache=cache)
                for key in keys_.T
            ]
        else:
            left = list(reversed(numpy.array(left).T**keys_.T))
        if isinstance(right, Dist):
            right = [
                evaluation.evaluate_moment(right, key, cache=cache)
                for key in keys_.T
            ]
        else:
            right = list(reversed(numpy.array(right).T**keys_.T))

        out = numpy.zeros(keys.shape)
        for idx in range(keys_.shape[1]):
            key = keys_.T[idx]
            coef = comb(keys.T, key)
            out += coef*left[idx]*right[idx]*(key <= keys.T)

        if len(self) > 1:
            out = numpy.prod(out, 1)
        return out

    def _ttr(self, kloc, left, right, cache):
        """
        Three terms recursion coefficients.

        Example:
            >>> print(numpy.around(chaospy.Uniform().ttr([0, 1, 2, 3]), 4))
            [[ 0.5     0.5     0.5     0.5   ]
             [-0.      0.0833  0.0667  0.0643]]
            >>> print(numpy.around(chaospy.Add(chaospy.Uniform(), 2).ttr([0, 1, 2, 3]), 4))
            [[ 2.5     2.5     2.5     2.5   ]
             [-0.      0.0833  0.0667  0.0643]]
            >>> print(numpy.around(chaospy.Add(2, chaospy.Uniform()).ttr([0, 1, 2, 3]), 4))
            [[ 2.5     2.5     2.5     2.5   ]
             [-0.      0.0833  0.0667  0.0643]]
            >>> print(numpy.around(chaospy.Add(1, 1).ttr([0, 1, 2, 3]), 4)) # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
                ...
            chaospy.distributions.baseclass.StochasticallyDependentError: recurrence ...
        """
        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise StochasticallyDependentError(
                    "sum of distributions not feasible: "
                    "{} and {}".format(left, right)
                )
        else:
            if not isinstance(right, Dist):
                raise StochasticallyDependentError(
                    "recurrence coefficients for constants not feasible: "
                    "{}".format(left+right)
                )
            left, right = right, left

        coeff0, coeff1 = evaluation.evaluate_recurrence_coefficients(
            left, kloc, cache=cache)
        return coeff0 + numpy.asarray(right), coeff1

    def __str__(self):
        if self._repr is None:
            return (self.__class__.__name__ + "(" + str(self.prm["left"]) +
                    ", " + str(self.prm["right"]) + ")")
        return super(Add, self).__str__()

    def _fwd_cache(self, cache):
        left = evaluation.get_forward_cache(self.prm["left"], cache)
        right = evaluation.get_forward_cache(self.prm["right"], cache)
        if not isinstance(left, Dist) and not isinstance(right, Dist):
            return left+right
        return self

    def _inv_cache(self, cache):
        left = evaluation.get_forward_cache(self.prm["left"], cache)
        right = evaluation.get_forward_cache(self.prm["right"], cache)
        if not isinstance(left, Dist) and not isinstance(right, Dist):
            return left+right
        return self


def add(left, right):
    return Add(left, right)



"""Weibull Distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class weibull(Dist):
    """Weibull Distribution."""

    def __init__(self, a=1):
        Dist.__init__(self, a=a)

    def _pdf(self, x, a):
        return a*x**(a-1)*numpy.e**(-x**a)

    def _cdf(self, x, a):
        return (1-numpy.e**(-x**a))

    def _ppf(self, q, a):
        return (-numpy.log(1-q+1*(q==1)))**(1./a)*(q!=1) + 30.**(1./a)*(q==1)

    def _mom(self, k, a):
        return special.gamma(1.+k*1./a)

    def _lower(self, a):
        return 0.

    def _upper(self, a):
        return 30.**(1./a)


class Weibull(Add):
    """
    Weibull Distribution

    Args:
        shape (float, Dist):
            Shape parameter.
        scale (float, Dist):
            Scale parameter.
        shift (float, Dist):
            Location of lower bound.

    Examples:
        >>> distribution = chaospy.Weibull(2)
        >>> distribution
        Weibull(scale=1, shape=2, shift=0)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([0.4724, 0.7147, 0.9572, 1.2686])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.7558, 0.8577, 0.7658, 0.5075])
        >>> distribution.sample(4).round(4)
        array([1.0296, 0.3495, 1.7325, 0.8113])
        >>> distribution.mom(1).round(4)
        0.8862
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=weibull(shape)*scale, right=shift)



"""Hyperbolic secant distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class hyperbolic_secant(Dist):
    """Hyperbolic secant distribution."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return .5*numpy.cosh(numpy.pi*x/2.)**-1

    def _cdf(self, x):
        return 2/numpy.pi*numpy.arctan(numpy.e**(numpy.pi*x/2.))

    def _ppf(self, q):
        return 2/numpy.pi*numpy.log(numpy.tan(numpy.pi*q/2.))

    def _mom(self, k):
        shape = k.shape
        output = numpy.abs([special.euler(k_)[-1] for k_ in k.flatten()])
        return output.reshape(shape)


class HyperbolicSecant(Add):
    """
    Hyperbolic secant distribution

    Args:
        loc (float, Dist):
            Location parameter
        scale (float, Dist):
            Scale parameter

    Examples:
        >>> distribution = chaospy.HyperbolicSecant(2, 2)
        >>> distribution
        HyperbolicSecant(loc=2, scale=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([0.5687, 1.5933, 2.4067, 3.4313])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.1469, 0.2378, 0.2378, 0.1469])
        >>> distribution.sample(4).round(4)
        array([ 2.6397, -0.1648,  5.2439,  1.9287])
        >>> distribution.mom(1).round(4)
        2.0
    """

    def __init__(self, loc=0, scale=1):
        self._repr = {"loc": loc, "scale": scale}
        Add.__init__(self, left=hyperbolic_secant()*scale, right=loc)



"""Pareto type 1 distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class pareto1(Dist):
    """Pareto type 1 distribution."""

    def __init__(self, b):
        Dist.__init__(self, b=b)

    def _pdf(self, x, b):
        return b * x**(-b-1)

    def _cdf(self, x, b):
        return 1 -  x**(-b)

    def _ppf(self, q, b):
        return pow(1-q, -1.0/b)

    def _lower(self, b):
        return 1.0


class Pareto1(Add):
    """
    Pareto type 1 distribution.

    Lower threshold at scale+loc and survival: x^-shape

    Args:
        shape (float, Dist):
            Tail index parameter
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.Pareto1(2, 2, 2)
        >>> distribution
        Pareto1(loc=2, scale=2, shape=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([4.2361, 4.582 , 5.1623, 6.4721])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.7155, 0.4648, 0.253 , 0.0894])
        >>> distribution.sample(4).round(4)
        array([ 5.3981,  4.126 , 10.9697,  4.7794])
    """

    def __init__(self, shape=1, scale=1, loc=0):
        self._repr = {"shape": shape, "scale": scale, "loc": loc}
        Add.__init__(self, left=pareto1(shape)*scale, right=loc)



"""Nakagami-m distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class nakagami(Dist):
    """Nakagami-m distribution."""

    def __init__(self, nu):
        Dist.__init__(self, nu=nu)

    def _pdf(self, x, nu):
        return 2*nu**nu/special.gamma(nu)*(x**(2*nu-1.0))*numpy.exp(-nu*x*x)

    def _cdf(self, x, nu):
        return special.gammainc(nu,nu*x*x)

    def _ppf(self, q, nu):
        return numpy.sqrt(1.0/nu*special.gammaincinv(nu, q))

    def _lower(self, nu):
        return 0.


class Nakagami(Add):
    """
    Nakagami-m distribution.

    Args:
        shape (float, Dist):
            Shape parameter
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.Nakagami(2, 2, 2)
        >>> distribution
        Nakagami(scale=2, shape=2, shift=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([3.284 , 3.6592, 4.0111, 4.4472])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.4642, 0.5766, 0.5383, 0.3669])
        >>> distribution.sample(4).round(4)
        array([4.1137, 3.076 , 5.0824, 3.8012])
        >>> distribution.mom(1).round(4)
        3.88
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, nakagami(shape)*scale, shift)



"""Logarithm of another distribution."""
import numpy

from ..baseclass import Dist
from .unary import UnaryOperator


class Logn(UnaryOperator):
    """
    Logarithm with base N.

    Args:
        dist (Dist):
            Distribution to perform transformation on.
        base (int, float):
            the logarithm base.

    Example:
        >>> distribution = chaospy.Logn(chaospy.Uniform(1, 2), 3)
        >>> print(distribution)
        Logn(Uniform(lower=1, upper=2), 3)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.166  0.3063 0.4278 0.535 ]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [1.3183 1.5381 1.7578 1.9775]
        >>> print(numpy.around(distribution.sample(4), 4))
        [0.4578 0.0991 0.608  0.3582]
        >>> print(numpy.around(distribution.mom(1), 4))
        0.3516
    """

    def __init__(self, dist, base=2):
        assert isinstance(dist, Dist)
        assert numpy.all(dist.lower > 0)
        assert base > 0 and base != 1
        Dist.__init__(self, dist=dist, base=base)
        self._repr = {"_": [dist, base]}

    def _post_pdf(self, xloc, base):
        return base**xloc*numpy.log(base)

    def _pre_fwd(self, xloc, base):
        return base**xloc

    def _post_fwd(self, uloc, base):
        return uloc

    def _pre_inv(self, qloc, base):
        return qloc

    def _post_inv(self, uloc, base):
        return numpy.log(uloc)/numpy.log(base)


class Log(Logn):
    """
    Logarithm with base Euler's constant.

    Args:
        dist (Dist):
            Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Log(chaospy.Uniform(1, 2))
        >>> print(distribution)
        Log(Uniform(lower=1, upper=2))
    """

    def __init__(self, dist):
        super(Log, self).__init__(dist=dist, base=numpy.e)
        self._repr = {"_": [dist]}


class Log10(Logn):
    """
    Logarithm with base 10.

    Args:
        dist (Dist): Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Log10(chaospy.Uniform(1, 2))
        >>> print(distribution)
        Log10(Uniform(lower=1, upper=2))
    """

    def __init__(self, dist):
        super(Log10, self).__init__(dist=dist, base=10)
        self._repr = {"_": [dist]}



"""Tukey-lambda distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class tukey_lambda(Dist):
    """Tukey-lambda distribution."""

    def __init__(self, lam):
        Dist.__init__(self, lam=lam)

    def _pdf(self, x, lam):
        lam = numpy.zeros(x.shape) + lam
        output = numpy.zeros(x.shape)
        indices = (lam <= 0) | (numpy.abs(x)*lam < 1)
        lam = lam[indices]
        Fx = special.tklmbda(x[indices], lam)
        Px = 1/(Fx**(lam-1.0) + ((1-Fx))**(lam-1.0))
        output[indices] = Px
        return output

    def _cdf(self, x, lam):
        return special.tklmbda(x, lam)

    def _ppf(self, q, lam):
        output = numpy.zeros(q.shape)
        lam = numpy.zeros(q.shape) + lam
        indices = lam != 0
        q_ = q[indices]
        lam_ = lam[indices]
        output[indices] = (q_**lam_ - (1-q_)**lam_)/lam_
        q_ = q[~indices]
        output[~indices] = numpy.log(q_/(1-q_))
        return output


class TukeyLambda(Add):
    """
    Tukey-lambda distribution.

    Args:
        lam (float, Dist):
            Shape parameter
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.TukeyLambda(0, 2, 2)
        >>> distribution
        TukeyLambda(scale=2, shape=0, shift=2)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> distribution.inv(q).round(4)
        array([-1.2189,  0.6137,  2.    ,  3.3863,  5.2189])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.1667, 0.3333, 0.5   , 0.6667, 0.8333])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.0694, 0.1111, 0.125 , 0.1111, 0.0694])
        >>> distribution.sample(4).round(4)
        array([ 3.2697, -2.0812,  7.9008,  1.8575])
        >>> distribution.mom(1).round(4)
        2.0
    """

    def __init__(self, shape=0, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=tukey_lambda(shape)*scale, right=shift)



"""Gompertz distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class gompertz(Dist):
    """Gompertz distribution."""

    def __init__(self, c):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        ex = numpy.exp(x)
        return c*ex*numpy.exp(-c*(ex-1))

    def _cdf(self, x, c):
        return 1.0-numpy.exp(-c*(numpy.exp(x)-1))

    def _ppf(self, q, c):
        return numpy.log(1-1.0/c*numpy.log(1-q))

    def _lower(self, c):
        return 0.


class Gompertz(Add):
    """
    Gompertz distribution

    Args:
        shape (float, Dist):
            Shape parameter
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.Gompertz(3, 2, 2)
        >>> distribution
        Gompertz(scale=2, shape=3, shift=2)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([2.1435, 2.3145, 2.5331, 2.859 ])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([1.2893, 1.0532, 0.7833, 0.4609])
        >>> distribution.sample(4).round(4)
        array([2.6052, 2.0798, 3.3868, 2.3967])
        >>> distribution.mom(1).round(4)
        2.5242
    """

    def __init__(self, shape, scale, shift):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=gompertz(shape)*scale, right=shift)



"""Beta distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class beta_(Dist):

    def __init__(self, a=1, b=1):
        Dist.__init__(self, a=a, b=b)

    def _pdf(self, x, a, b):
        return x**(a-1)*(1-x)**(b-1)/ \
            special.beta(a, b)

    def _cdf(self, x, a, b):
        return special.btdtr(a, b, x)

    def _ppf(self, q, a, b):
        return special.btdtri(a, b, q)

    def _mom(self, k, a, b):
        return special.beta(a+k,b)/special.beta(a,b)

    def _ttr(self, n, a, b):
        nab = 2*n+a+b
        A = ((a-1)**2-(b-1)**2)*.5/\
                (nab*(nab-2) + (nab==0) + (nab==2)) + .5
        B1 = a*b*1./((a+b+1)*(a+b)**2)
        B2 = (n+a-1)*(n+b-1)*n*(n+a+b-2.)/\
            ((nab-1)*(nab-3)*(nab-2)**2+2.*((n==0)+(n==1)))
        B = numpy.where((n==0)+(n==1), B1, B2)
        return A, B

    def _lower(self, a, b):
        return 0.

    def _upper(self, a, b):
        return 1.


class Beta(Add):
    R"""
    Beta Probability Distribution.

    Args:
        alpha (float, Dist):
            First shape parameter, alpha > 0
        beta (float, Dist):
            Second shape parameter, b > 0
        lower (float, Dist):
            Lower threshold
        upper (float, Dist):
            Upper threshold

    Examples:
        >>> distribution = chaospy.Beta(2, 2, 2, 3)
        >>> distribution
        Beta(alpha=2, beta=2, lower=2, upper=3)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([2.2871, 2.4329, 2.5671, 2.7129])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([1.2281, 1.473 , 1.473 , 1.2281])
        >>> distribution.sample(4).round(4)
        array([2.6039, 2.2112, 2.8651, 2.4881])
        >>> distribution.mom(1).round(4)
        2.5
        >>> distribution.ttr([1, 2, 3]).round(4)
        array([[2.5   , 2.5   , 2.5   ],
               [0.05  , 0.0571, 0.0595]])
    """

    def __init__(self, alpha, beta, lower=0, upper=1):
        self._repr = {
            "alpha": alpha, "beta": beta, "lower": lower, "upper": upper}
        Add.__init__(self, left=beta_(alpha, beta)*(upper-lower), right=lower)


class ArcSinus(Add):
    """
    Generalized Arc-sinus distribution

    Args:
        shape (float, Dist):
            Shape parameter where 0.5 is the default non-generalized case.
            Defined on the interval ``[0, 1]``.
        lower (float, Dist):
            Lower threshold
        upper (float, Dist):
            Upper threshold

    Examples:
        >>> distribution = chaospy.ArcSinus(0.8, 4, 6)
        >>> distribution
        ArcSinus(lower=4, shape=0.8, upper=6)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> distribution.inv(q).round(4)
        array([4.9875, 5.6438, 5.9134, 5.9885, 5.9996])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.1667, 0.3333, 0.5   , 0.6667, 0.8333])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([ 0.1857,  0.3868,  1.1633,  5.8145, 92.8592])
        >>> distribution.sample(4).round(4)
        array([5.9861, 4.6882, 6.    , 5.897 ])
        >>> distribution.mom(1)
        array(5.6)
    """

    def __init__(self, shape=0.5, lower=0, upper=1):
        self._repr = {"shape": shape, "lower": lower, "upper": upper}
        Add.__init__(
            self, left=beta_(shape, 1-shape)*(upper-lower), right=lower)


class PowerLaw(Add):
    """
    Powerlaw distribution

    Args:
        shape (float, Dist):
            Shape parameter
        lower (float, Dist):
            Location of lower threshold
        upper (float, Dist):
            Location of upper threshold

    Examples:
        >>> distribution = chaospy.PowerLaw(0.8, 4, 6)
        >>> distribution
        PowerLaw(lower=4, shape=0.8, upper=6)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> distribution.inv(q).round(4)
        array([4.213 , 4.5066, 4.8409, 5.2048, 5.5924])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.1667, 0.3333, 0.5   , 0.6667, 0.8333])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.626 , 0.5264, 0.4757, 0.4427, 0.4187])
        >>> distribution.sample(4).round(4)
        array([5.1753, 4.1339, 5.8765, 4.8036])
        >>> distribution.mom(1).round(4)
        4.8889
    """

    def __init__(self, shape=1, lower=0, upper=1):
        self._repr = {"shape": shape, "lower": lower, "upper": upper}
        Add.__init__(self, left=beta_(shape, 1)*(upper-lower), right=lower)


class Wigner(Add):
    """
    Wigner (semi-circle) distribution

    Args:
        radius (float, Dist):
            Radius of the semi-circle (scale)
        shift (float, Dist):
            Location of the circle origin (location)

    Examples:
        >>> distribution = chaospy.Wigner(2, 3)
        >>> distribution
        Wigner(radius=2, shift=3)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> distribution.inv(q).round(4)
        array([1.8934, 2.4701, 3.    , 3.5299, 4.1066])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.1667, 0.3333, 0.5   , 0.6667, 0.8333])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.2651, 0.3069, 0.3183, 0.3069, 0.2651])
        >>> distribution.sample(4).round(4)
        array([3.4874, 1.6895, 4.6123, 2.944 ])
        >>> distribution.mom(1).round(4)
        3.0
    """

    def __init__(self, radius=1, shift=0):
        self._repr = {"radius": radius, "shift": shift}
        Add.__init__(self, left=radius*(2*beta_(1.5, 1.5)-1), right=shift)


class PERT(Beta):
    r"""
    Program Evaluation and Review Technique (PERT) Distribution.

    Defined by its mean::

        \mu = \frac{lower + gamma*mode + upper}{2 + gamma}

    Normal PERT for `gamma=4`. Other values results in the so called
    modified-PERT distribution.

    Args:
        lower (float):
            The lower bounds for the distribution.
        mode (float, Dist):
            The mode of the distribution.
        upper (float):
            The upper bounds for the distribution.
        gamma (flat, Dist):
            Modify the PERT distribution to make more emphasis on the
            distribution mode instead of the distribution tails.

    Examples:
        >>> distribution = chaospy.PERT(-1, 0, 1)
        >>> distribution
        PERT(gamma=4, lower=-1, mode=0, upper=1)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> distribution.inv(q).round(4)
        array([-0.3946, -0.1817,  0.    ,  0.1817,  0.3946])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.1667, 0.3333, 0.5   , 0.6667, 0.8333])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.6683, 0.8766, 0.9375, 0.8766, 0.6683])
        >>> distribution.sample(4).round(4)
        array([ 0.1669, -0.4788,  0.6223, -0.019 ])
        >>> distribution.mom(1).round(4)
        0.0
    """

    def __init__(self, lower, mode, upper, gamma=4):
        mu = (lower+4*mode+upper)/6.
        alpha = 1+gamma*(mu-lower)/(upper-lower)
        beta = 1+gamma*(upper-mu)/(upper-lower)
        Beta.__init__(self, alpha=alpha, beta=beta, lower=lower, upper=upper)
        self._repr = {"lower": lower, "mode": mode, "upper": upper, "gamma": gamma}



"""Log-uniform distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add


class log_uniform(Dist):
    """Log-uniform distribution."""

    def __init__(self, lo=0, up=1):
        Dist.__init__(self, lo=lo, up=up)

    def _pdf(self, x, lo, up):
        return 1./(x*(up-lo))

    def _cdf(self, x, lo, up):
        return (numpy.log(x)-lo)/(up-lo)

    def _ppf(self, q, lo, up):
        return numpy.e**(q*(up-lo) + lo)

    def _lower(self, lo, up):
        return numpy.e**lo

    def _upper(self, lo, up):
        return numpy.e**up

    def _mom(self, k, lo, up):
        return ((numpy.e**(up*k)-numpy.e**(lo*k))/((up-lo)*(k+(k==0))))**(k!=0)


class LogUniform(Add):
    """
    Log-uniform distribution

    Args:
        lower (float, Dist):
            Location of lower threshold of uniform distribution.
        upper (float, Dist):
            Location of upper threshold of uniform distribution.
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.LogUniform(2, 3, 2, 3)
        >>> distribution
        LogUniform(lower=2, scale=2, shift=3, upper=3)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([21.05  , 25.0464, 29.9275, 35.8893])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.0554, 0.0454, 0.0371, 0.0304])
        >>> distribution.sample(4).round(4)
        array([31.4099, 19.5793, 41.2227, 26.9349])
        >>> distribution.mom(1).round(4)
        28.393
    """

    def __init__(self, lower=0, upper=1, scale=1, shift=0):
        self._repr = {"lower": lower, "upper": upper, "scale": scale, "shift": shift}
        Add.__init__(self, left=log_uniform(lower, upper)*scale, right=shift)



"""Cauchy distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add


class cauchy(Dist):
    """Standard Cauchy distribution."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return 1.0/numpy.pi/(1.0+x*x)

    def _cdf(self, x):
        return 0.5 + 1.0/numpy.pi*numpy.arctan(x)

    def _ppf(self, q):
        return numpy.tan(numpy.pi*q-numpy.pi/2.0)


class Cauchy(Add):
    """
    Cauchy distribution.

    Args:
        loc (float, Dist):
            Location parameter
        scale (float, Dist):
            Scaling parameter

    Examples:
        >>> distribution = chaospy.Cauchy(2, 4)
        >>> distribution
        Cauchy(loc=2, scale=4)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> distribution.inv(q).round(4)
        array([-4.9282, -0.3094,  2.    ,  4.3094,  8.9282])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.1667, 0.3333, 0.5   , 0.6667, 0.8333])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.0199, 0.0597, 0.0796, 0.0597, 0.0199])
        >>> distribution.sample(4).round(4)
        array([ 4.0953, -8.585 , 27.4011,  1.776 ])
        >>> distribution.mom(1).round(4)
        2.0
    """

    def __init__(self, loc=0, scale=1):
        self._repr = {"loc": loc, "scale": scale}
        Add.__init__(self, left=cauchy()*scale, right=loc)



"""Kurtosis operator."""
import numpy
import numpoly

from .. import distributions
from .expected import E
from .standard_deviation import Std


def Kurt(poly, dist=None, fisher=True, **kws):
    """
    Kurtosis operator.

    Element by element 4rd order statistics of a distribution or polynomial.

    Args:
        poly (numpoly.ndpoly, Dist):
            Input to take kurtosis on.
        dist (Dist):
            Defines the space the skewness is taken on. It is ignored if
            ``poly`` is a distribution.
        fisher (bool):
            If True, Fisher's definition is used (Normal -> 0.0). If False,
            Pearson's definition is used (normal -> 3.0)

    Returns:
        (numpy.ndarray):
            Element for element variance along ``poly``, where
            ``skewness.shape==poly.shape``.

    Examples:
        >>> dist = chaospy.J(chaospy.Gamma(1, 1), chaospy.Normal(0, 2))
        >>> chaospy.Kurt(dist).round(4)
        array([6., 0.])
        >>> chaospy.Kurt(dist, fisher=False).round(4)
        array([9., 3.])
        >>> q0, q1 = chaospy.variable(2)
        >>> poly = chaospy.polynomial([1, q0, q1, 10*q0*q1-1])
        >>> chaospy.Kurt(poly, dist).round(4)
        array([nan,  6.,  0., 15.])
    """
    adjust = 3 if fisher else 0

    if dist is None:
        dist, poly = poly, numpoly.variable(len(poly))
    poly = numpoly.set_dimensions(poly, len(dist))
    if not poly.isconstant:
        return poly.tonumpy()**4-adjust

    poly = poly-E(poly, dist, **kws)
    poly = numpoly.true_divide(poly, Std(poly, dist, **kws))
    return E(poly**4, dist, **kws)-adjust



"""Reciprocal distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add


class Reciprocal(Dist):
    """
    Reciprocal distribution.

    Args:
        lower (float, Dist):
            Lower threshold of distribution. Must be smaller than ``upper``.
        upper (float, Dist):
            Upper threshold of distribution.

    Examples:
        >>> distribution = chaospy.Reciprocal(2, 4)
        >>> distribution
        Reciprocal(lower=2, upper=4)
        >>> q = numpy.linspace(0, 1, 5)
        >>> distribution.inv(q).round(4)
        array([2.    , 2.3784, 2.8284, 3.3636, 4.    ])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.  , 0.25, 0.5 , 0.75, 1.  ])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.7213, 0.6066, 0.5101, 0.4289, 0.3607])
        >>> distribution.sample(4).round(4)
        array([3.1462, 2.166 , 3.8645, 2.7937])
        >>> distribution.mom(1).round(4)
        7.8433
    """

    def __init__(self, lower=1, upper=2):
        self._repr = {"lower": lower, "upper": upper}
        Dist.__init__(self, lower=lower, upper=upper)

    def _pdf(self, x, lower, upper):
        return 1./(x*numpy.log(upper/lower))

    def _cdf(self, x, lower, upper):
        return numpy.log(x/lower)/numpy.log(upper/lower)

    def _ppf(self, q, lower, upper):
        return numpy.e**(q*numpy.log(upper/lower) + numpy.log(lower))

    def _lower(self, lower, upper):
        return lower

    def _upper(self, lower, upper):
        return upper

    def _mom(self, k, lower, upper):
        return ((upper*numpy.e**k-lower*numpy.e**k)/(numpy.log(upper/lower)*(k+(k == 0))))**(k != 0)



"""Bradford distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add


class bradford(Dist):
    """Standard Bradford distribution."""

    def __init__(self, c=1):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return  c / (c*x + 1.0) / numpy.log(1.0+c)

    def _cdf(self, x, c):
        return numpy.log(1.0+c*x) / numpy.log(c+1.0)

    def _ppf(self, q, c):
        return ((1.0+c)**q-1)/c

    def _lower(self, c):
        return 0

    def _upper(self, c):
        return 1


class Bradford(Add):
    """
    Bradford distribution.

    Args:
        shape (float, Dist):
            Shape parameter
        lower (float, Dist):
            Location of lower threshold
        upper (float, Dist):
            Location of upper threshold

    Examples:
        >>> distribution = chaospy.Bradford(0.8, 4, 6)
        >>> distribution
        Bradford(lower=4, shape=0.8, upper=6)
        >>> q = numpy.linspace(0, 1, 5)
        >>> distribution.inv(q).round(4)
        array([4.    , 4.3957, 4.8541, 5.385 , 6.    ])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.  , 0.25, 0.5 , 0.75, 1.  ])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.6805, 0.5875, 0.5072, 0.4379, 0.3781])
        >>> distribution.sample(4).round(4)
        array([5.171 , 4.1748, 5.8704, 4.8192])
        >>> distribution.mom(1).round(4)
        4.9026
    """
    def __init__(self, shape=1, lower=0, upper=1):
        self._repr = {"shape": shape, "lower": lower, "upper": upper}
        Add.__init__(self, left=bradford(shape)*(upper-lower), right=lower)



"""
Evaluate distribution bounds: The location where it is reasonable to define the
distribution density is located between.

Example usage
-------------

Define a simple distribution and data::

    >>> class Uniform(chaospy.Dist):
    ...     _lower = lambda self, lo, up: lo
    ...     _upper = lambda self, lo, up: up
    >>> dist = Uniform(lo=1, up=3)

Normal usage::

    >>> evaluate_lower(dist)
    array([1.])
    >>> evaluate_upper(dist)
    array([3.])
    >>> evaluate_lower(dist, parameters={"lo": -1.})
    array([-1.])
"""
import numpy
from .parameters import load_parameters


def evaluate_lower(
        distribution,
        parameters=None,
        cache=None,
):
    """
    Evaluate lower and upper bounds.

    Args:
        distribution (Dist):
            Distribution to evaluate.
        parameters (:py:data:typing.Any):
            Collection of parameters to override the default ones in the
            distribution.
        cache (:py:data:typing.Any):
            A collection of previous calculations in case the same distribution
            turns up on more than one occasion.

    Returns:
        The lower and upper bounds of ``distribution`` at location
        ``x_data`` using parameters ``parameters``.
    """
    cache = cache if cache is not None else {}

    parameters = load_parameters(
        distribution, "_lower", parameters=parameters)

    lower = distribution._lower(**parameters)
    lower = numpy.asfarray(lower)+numpy.zeros(len(distribution))

    cache[distribution] = lower
    return lower


def evaluate_upper(
        distribution,
        parameters=None,
        cache=None,
):
    """
    Evaluate lower and upper bounds.

    Args:
        distribution (Dist):
            Distribution to evaluate.
        parameters (:py:data:typing.Any):
            Collection of parameters to override the default ones in the
            distribution.
        cache (:py:data:typing.Any):
            A collection of previous calculations in case the same distribution
            turns up on more than one occasion.

    Returns:
        The lower and upper bounds of ``distribution`` at location
        ``x_data`` using parameters ``parameters``.
    """
    cache = cache if cache is not None else {}

    parameters = load_parameters(
        distribution, "_upper", parameters=parameters)

    upper = distribution._upper(**parameters)
    upper = numpy.asfarray(upper)+numpy.zeros(len(distribution))
    cache[distribution] = upper
    return upper



"""Burr Type III distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class burr(Dist):
    """Stadard Burr distribution."""

    def __init__(self, alpha=1., kappa=1.):
        Dist.__init__(self, alpha=alpha, kappa=kappa)

    def _pdf(self, x, alpha, kappa):
        output = numpy.zeros(x.shape)
        indices = x > 0
        output[indices] = alpha*kappa*x[indices]**(-alpha-1.0)
        output[indices] /= (1+x[indices]**alpha)**(kappa+1)
        return output

    def _cdf(self, x, alpha, kappa):
        output = numpy.zeros(x.shape)
        indices = x > 0
        output[indices] = 1-(1+x[indices]**alpha)**-kappa
        return output

    def _ppf(self, q, alpha, kappa):
        return ((1-q)**(-1./kappa) -1)**(1./alpha)

    def _mom(self, k, alpha, kappa):
        return kappa*special.beta(1-k*1./alpha, kappa+k*1./alpha)

    def _lower(self, alpha, kappa):
        return 0.


class Burr(Add):
    """
    Burr Type XII or Singh-Maddala distribution.

    Args:
        alpha (float, Dist):
            First shape parameter
        kappa (float, Dist):
            Second shape parameter
        loc (float, Dist):
            Location parameter
        scale (float, Dist):
            Scaling parameter

    Examples:
        >>> distribution = chaospy.Burr(100, 1.2, 4, 2)
        >>> distribution
        Burr(alpha=100, kappa=1.2, loc=4, scale=2)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> distribution.inv(q).round(4)
        array([5.9642, 5.9819, 5.9951, 6.0081, 6.0249])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.1667, 0.3333, 0.5   , 0.6667, 0.8333])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([266.5437,  71.6255,  21.5893,   5.3229,   0.643 ])
        >>> distribution.sample(4).round(4)
        array([6.007 , 5.9558, 6.0489, 5.9937])
        >>> distribution.mom(1).round(4)
        6.0061
    """

    def __init__(self, alpha=1, kappa=1, loc=0, scale=1):
        self._repr = {
            "alpha": alpha, "kappa": kappa, "loc": loc, "scale": scale}
        Add.__init__(self, left=burr(alpha, kappa)*scale, right=loc)



"""Log-gamma distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class log_gamma(Dist):
    """Log-gamma distribution."""

    def __init__(self, c):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return numpy.exp(c*x-numpy.exp(x)-special.gammaln(c))

    def _cdf(self, x, c):
        return special.gammainc(c, numpy.exp(x))

    def _ppf(self, q, c):
        return numpy.log(special.gammaincinv(c,q))


class LogGamma(Add):
    """
    Log-gamma distribution

    Args:
        shape (float, Dist):
            Shape parameter
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.LogGamma(2, 2, 1)
        >>> distribution
        LogGamma(scale=2, shape=2, shift=1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([0.6138, 1.639 , 2.4085, 3.1934])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.149 , 0.2392, 0.2706, 0.2245])
        >>> distribution.sample(4).round(4)
        array([ 2.6074, -0.0932,  4.1166,  1.9675])
        >>> distribution.mom(1).round(4)
        1.8456
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=log_gamma(shape)*scale, right=shift)



"""Folded Cauchy distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add


class folded_cauchy(Dist):
    """Folded Cauchy distribution."""

    def __init__(self, c=0):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return 1./(numpy.pi*(1+(x-c)**2)) + 1./(numpy.pi*(1+(x+c)**2))

    def _cdf(self, x, c):
        return (numpy.arctan(x-c) + numpy.arctan(x+c))/numpy.pi

    def _lower(self, c):
        return 0.

    def _upper(self, c):
        return 10.**10  # actually infinity


class FoldedCauchy(Add):
    """
    Folded Cauchy distribution.

    Args:
        shape (float, Dist):
            Shape parameter
        scale (float, Dist):
            Scaling parameter
        shift (float, Dist):
            Location parameter

    Examples:
        >>> distribution = chaospy.FoldedCauchy(3, 2, 1)
        >>> distribution
        FoldedCauchy(scale=2, shape=3, shift=1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> distribution.inv(q).round(4)
        array([ 5.1449,  6.708 ,  8.0077, 10.6502])
        >>> distribution.fwd(distribution.inv(q)).round(4)
        array([0.2, 0.4, 0.6, 0.8])
        >>> distribution.pdf(distribution.inv(q)).round(4)
        array([0.0915, 0.1603, 0.1306, 0.0393])
        >>> distribution.sample(4).round(4)
        array([9.2648, 7.5918, 7.2159, 7.1782])
    """

    def __init__(self, shape=0, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=folded_cauchy(shape)*scale, right=shift)
